;-----------------------------------------------------------------------------
;* lcd2313.asm               letzte Aenderung: 18.4.2010
;* Version: 0.0
;* Autor: Rolf Pfister
;* Copyright: Freeware
;* Prozessor: ATtiny2313A, 8MHz interner Oszillator
;* Schaltschema:
;*	 PortB: B2-B7 verbunden mit LCD
;*	 PortD:	D0 Eingabepin fuer Daten von Anwendung
;*
;* History:
;* 14.4.2010	Erstellung
;*
; Globale Registerbelegung:
; r8: Sicherung von Status im Timer-Interrupt
; r9: Zaehlung Microsec
; r10-r15: Millisek bis Tage
;-----------------------------------------------------------------------------
.include "tn2313Adef.inc" 	; Definitionsdatei fuer den Prozessortyp
.equ SRAMSTART=0x60	;beim ATtiny2313A

;.define DEBUGGING
;.define TESTCODE
;.define FREQTEST
;.define TIMERTEST
.define TEST1

;-----------------------------------------------------------------------------
; Reset und Interruptvectoren	;VNr. Beschreibung
begin:	rjmp	main		; 0   Power On Reset
	reti			; 1   Int0-Interrupt
	reti			; 2   Int1-Interrupt
	reti			; 3   TC1 Capture
	reti			; 4   TC1 Compare Match A
	reti			; 5   TC1 Overflow
	rjmp tc0over		; 6   TC0 Overflow
	reti			; 7   UART Rx Complete
	reti			; 8   UART Data Register Empty
	reti			; 9   UART Tx Complete
	reti			; 10  Analog Comperator
	reti			; 11  Pin Change B
	reti			; 12  OC1B
	reti			; 13  OC0A
	reti			; 14  OC0B
	reti			; 15  USI Start Condition
	reti			; 16  USI Overflow
	reti			; 17  ERDY
	reti			; 18  Watchdog Timer Overflow
	reti			; 19  Pin Change A
	reti			; 20  Pin Change D

;-----------------------------------------------------------------------------
; Variablen im SRAM:
.equ lcdpos=SRAMSTART		;LCD-Position
.equ sollH=SRAMSTART+1
.equ sollL=SRAMSTART+2		;Sollwert fuer Pulslaengen-Vergleich
.ifdef TESTCODE
.equ lcdn0=SRAMSTART+3
.equ lcdn1=SRAMSTART+4
.equ lcdn2=SRAMSTART+5
.equ lcdn3=SRAMSTART+6		;Zahlen fuers LCD
.endif

;-----------------------------------------------------------------------------
; Start, Power On, Reset
main:	ldi r16, RAMEND
        out SPL, r16  	     ; Init Stackpointer L

;; Init-Code
	ldi r16, 0xFF		;Alle 8 Leitungen auf Ausgang
;	ldi r16, 0xFC		;Obere 6 Leitungen auf Ausgang
        out DDRB, r16		;Data Direction Register B setzen
	ldi r16, 0
	out PORTB, r16		;alle aus
	ldi r17, 0x01		;Pullup fuer Engangspin
	out PORTD, r17		;Pullup-Widerstand an D0

	ldi r16, 0x80
	out CLKPR, r16		; Vorbereitung zum Prescaler setzen
	ldi r16, 0
	out CLKPR, r16		; Clk-Prescaler auf 1 setzen --> 8MHz

.ifdef FREQTEST
t1:	inc r16
	out PORTB, r16		;Schlaufe braucht 4 Takte, ganze Periode also 8
	rjmp t1			;An PB0 gemessene Frequenz*8 = Systemtakt
.endif

	ldi r17, (1<<TOIE0)
	out TIMSK, r17		;Timer/Counter0 Overflow einschalten
	ldi r17, 3		;FastPWM-Modus
	out TCCR0A, r17
	ldi r17, 2+(1<<WGM02)	;1=nicht teilen, 2=durch 8, 3=64, 4=256, 5=1024
	out TCCR0B, r17		;Vorteiler setzen
;Bei 8MHz: Schrittweite 1 usec
;Um alle 10usec einen Interrupt zu erhalten jeweils auf 10 zaehlen
	ldi r17, 10-1		;1 weniger weil 0 auch 1 Takt braucht
	out OCR0A, r17
	ldi r17, 0
	out TCNT0, r17		;bei 0 beginnen

	ldi r16, 100
	mov r9, r16		; Microsekunden mit Schrittweite 10
	clr r10
	clr r11			; r11r10 = Millisec
	clr r12			; Sekunden
	clr r13			; Minuten
	clr r14			; Stunden
	clr r15			; Tage, auf 0 setzen
	sei			;Interrupt erlauben

.ifdef TIMERTEST
T1:	sbi PORTB, 0
	ldi r16, 1
	rcall milliwait1
	cbi PORTB, 0
	ldi r16, 1
	rcall milliwait1	;sollte 500Hz geben
	rjmp T1
.endif

	rcall lcd_init
.ifdef TEST1
test_1:	ldi r16, 100
	rcall milliwait1
	ldi r16, 0x11
	rcall lcd_go
	rcall print_hallo
	ldi r16, 0x21
	rcall lcd_go
	rcall print_hallo2
	ldi r16,  0x11
	rcall lcd_go
.endif

;-----------------------------------------------------------------------------
mainloop:
	rcall pulsmessen
	tst r21			;war Pulslange groesser 255ms ?
	brne err1
;	rjmp testpunkt1		;test
	asr r20			;halbe Pulslaenge berechnen
	brcc L4
	subi r22, -100
L4:	asr r22
	sts sollH, r20
	sts sollL, r22		;Sollpulslaengen fuer nachfolgende Bits
	clr r19
	rcall pulsmessen
	rcall pulsvergleich
	brcs L5			; langer Puls: 0
	inc r19			; kurzer PUls: 1
L5:	clr r18
	ldi r17, 8		; 8Bits empfangen
Lop2:	add r18, r18
	rcall pulsmessen
	rcall pulsvergleich
	brcs L6
	inc r18
L6:	dec r17
	brne Lop2
bytegelesen:
	mov r16, r18
	tst r19
	breq lcd_w
	rcall lcd_cmd
	rjmp L7
lcd_w:	rcall lcd_write
L7:	rjmp mainloop

err1:	rcall print_error		;Fehlerbehandlung (zu lange Pulse)
	rjmp mainloop
testpunkt1:
	rcall print_testpunkt1
	ldi r16, low(999)
	ldi r17, high(999)
	rcall milliwait
	rjmp mainloop

pulsmessen:			; Resultat in r21:r20.r22, r16 zerstoert
	push r17
	push r18
	push r19
L1:	in r16, PIND
	andi r16, 1
	brne L1
	cli
	mov r16, r9		; aktuelle Zeit r16 = usec
	mov r19, r11
	mov r18, r10		; r19:r18 = aktuelle msec
	sei
	ldi r17, 100		; vom Bereich 100 bis 1 auf
	sub r17, r16		; Bereich 0 bis 99 umrechnen.
L2:	in r16, PIND
	andi r16, 1
	breq L2			; Auf Ende des Startbits warten
	cli
	mov r16, r9		; aktuelle Zeit r16 = usec
	mov r21, r11
	mov r20, r10		; r21:r20 = aktuelle msec
	sei
	ldi r22, 100
	sub r22, r16		;Bereich 0 bis 99 entspricht 0.00 bis 0.99msec
	sub r22, r17
	brcc L3
	subi r22, -100		;bei negativem Ergebnis 100 dazuzaehlen
	sec			;und Cy setzen.
L3:	sbc r20, r18
	sbc r21, r19
	brcc fertig
	subi r20, low(-1000)
	sbci r21, high(-1000)	;bei negativem Ergebnis 1000 dazuzaehlen
fertig:	; Zeitdifferenz steht jetzt in r21:r20.r22
	pop r19
	pop r18
	pop r17
	ret

pulsvergleich:		; Cy: gesetzt wenn r21:r20.r22 groesser als soll
	tst r21
	breq v1
	sec
	ret
v1:	lds r16, sollH
	sub r16, r20
	breq v2
	ret
v2:	lds r16, sollL
	sub r16, r22
	ret
	
;-----------------------------------------------------------------------------
; Interrupt-Routine, sollte alle 10 usec kommen
; Registerbelegungen:
; r9: Microsekunden in 10usec Schritten von 100 auf 1 zaehlend
; r11:r10 Millisekunden, r12 Sekunden, r13 Minuten, r14 Stunden, r15 Tage
tc0over:push r16
	in r8, SREG	;Statusregister sichern
	dec r9
	breq tc1
tc0reti:
	out SREG, r8
	pop r16
	reti
tc1:	ldi r16, 100
	mov r9, r16
	sei
millisek:
	push r17
	ldi r16, 0xE8
	ldi r17, 0x03	;r17:r16 = 1000 = 0x03E8
	inc r10		;Millisekunden erhoehen
	brne mL1
	inc r11
mL1:	cp r10, r16
	cpc r11, r17
	brne end1	;fertig wenn msec!=1000
	clr r10
	clr r11		;Millisekunden auf 0 setzen
	inc r12		;Sekunden erhoehen
	ldi r16, 60
	cp r12, r16
	brne end1	;fertig wenn sec!=60
	clr r12		;Sekunden auf 0 setzen
	inc r13		;Minuten erhoehen
	cp r13, r16
	brne end1	;fertig wenn min!=60
	clr r13		;Minuten auf 0 setzen
	inc r14		;Stunden erhoehen
	ldi r16, 24
	cp r14, r16
	brne end1	;fertig wenn std!=24
	clr r14		;Stunden auf 0 setzen
	inc r15		;Tage erhoehen
end1:
	pop r17
	rjmp tc0reti

;-----------------------------------------------------------------------------
milliwait1:			;Wartezeit: r16, 1 bis 255 msec
	push r16
	push r17
	clr r17
	rjmp miL1
milliwait255:
	ldi r16, 255
milliwait:			;Wartezeit: r17:r16 msec, min 1 max 1000
	push r16
	push r17
miL1:	push r18
	push r19
	mov r18, r10
loop1:	cp  r18, r10
	breq loop1		;1. Millisekunde abwarten (0 bis 1 ms)
	mov r18, r17		; Oberstes Bit von r17 nach r18 kopieren
	andi r17, 0x7F		; und in r17 loeschen.
	add r16, r10
	adc r17, r11		;sollzeit = Wartezeit + aktuelle Millisek
	subi r16, low(1001)
	sbci r17, high(1001)	;sollzeit -= 1; sollzeit -= 1000;
	brpl loop2		;wenn sollzeit>=0 -->
	subi r16, low(-1000)	;sonst haben wir 1000 zuviel subtrahiert
	sbci r17, high(-1000)	;also wieder 1000 addieren (soll -= -1000)
loop2:	mov r19, r10
	subi r19, -8
	cp r16, r10
	cpc r17, r11		;auf sollzeit warten
	cpc r16, r10		;falls zwischen 1. u 2. cp ein Interrupt war
	brne loop2
pop19ret: pop r19
pop18ret: pop r18
pop17ret: pop r17
	  pop r16
	  ret

string_L1:
	rcall lcd_write
print_string:	    ;String in ZH:ZL auf lcd ausgeben, Z erhoeht, r16 zerst.
	lpm r16, Z+
	tst r16
	brne string_L1
	ret

.ifdef STACKERROR
Stack:	.db "Stack",0
print_stack:
	push ZL
	push ZH
	ldi r16, 'S'
	sts errorflag, r16
	ldi ZL, low(Stack<<1)
	ldi ZH, high(Stack<<1)
	rcall print_string
	pop ZH
	pop ZL
	ret
print_error17:			; r17 ausgeben und 'Error' auf LCD
	mov r16, r17
	rcall lcd_r16dez
	rjmp print_error
.endif
Error:	.db "Error",0
print_error:
	ldi r16, 'E'
;	sts errorflag, r16
	push ZL
	push ZH
	ldi ZL, low(Error<<1)
	ldi ZH, high(Error<<1)
	rcall print_string
	pop ZH
	pop ZL
	ret
Testp1:	.db "test1",0
print_testpunkt1:
	push ZL
	push ZH
	ldi ZL, low(Testp1<<1)
	ldi ZH, high(Testp1<<1)
	rcall print_string
	pop ZH
	pop ZL
	ret

.ifdef TEST1
Hallo:	.db "Hallo",0
print_hallo:
	push ZL
	push ZH
	ldi ZL, low(Hallo<<1)
	ldi ZH, high(Hallo<<1)
	rcall print_string
	pop ZH
	pop ZL
	ret
Hallo2:	.db "Zweite Zeile",0,0
print_hallo2:
	push ZL
	push ZH
	ldi ZL, low(Hallo2<<1)
	ldi ZH, high(Hallo2<<1)
	rcall print_string
	pop ZH
	pop ZL
	ret
.endif

;-----------------------------------------------------------------------------
;; Anzeige auf LCD-Modul
microwait:			; etwa r16 usec warten
	push r16
	rjmp mic_L1
microwait1:			; etwa 1 usec warten
	push r16
	push r17
	ldi r16, 4
	rjmp mic_L2
microwait50:
	push r16
	ldi r16, 50
mic_L1: push r17
	add r16, r16
	add r16, r16
mic_L2:	dec r16
	brne mic_L2
	pop r17
	pop r16
	ret

milliwait5:
	ldi r16, 5
	rjmp milliwait1

.ifdef TESTCODE
lcd_ziffer:		; void lcd_ziffer(int n) ;fuer 0<=n<=9
	ori r16, 0x30
	cpi r16, 0x3A	; oder fuer Hex-Ziffern bis n<=15
	brcs lcd_write
	subi r16, -7	; fuer Hex-Ziffern z.B. 3A --> 41=='A'
.endif
lcd_write:		;void lcd_write(char r16=c)
	sbi PORTB, 2
	rjmp lcd_send

lcd_cmd:	;void lcd_cmd(char cmd)
	cbi PORTB, 2
	rcall microwait1
lcd_send:	; lcd_send(char data) ;r16=data
	push r17
	push r18
	in r17, PORTB	;rs=PORTB
	andi r17, 4+3	;rs &= 4; (+3 zum Pullup-Widerstaende behalten)
	mov r18, r16	;char tmp=data
	andi r18, 0xF0	;tmp &= 0xF0
	or r18, r17	;tmp |= rs
	out PORTB, r18	;PORTB=tmp
	rcall microwait1
	sbi PORTB, 3	;PORTB |= (1<<3); //sbi(PORTB,3)
	rcall microwait1
	cbi PORTB, 3	;PORTB &= ~(1<<3); //cbi(PORTB,3)
	mov r18, r16	;tmp=data
	andi r18, 0x0F	;tmp &= 0x0F
	swap r18	;tmp <<= 4
	or r18, r17	;tmp |= rs
	out PORTB, r18	;PORTB=tmp
	rcall microwait1
	sbi PORTB, 3	;PORTB |= (1<<3); //sbi(PORTB,3)
	rcall microwait1
	cbi PORTB, 3	;PORTB &= ~(1<<3); //cbi(PORTB,3)
	ldi r16, 1
	rcall milliwait1
	pop r18
	pop r17
	ret

lcd_clear:
	push r16
	ldi r16, 0x01
	rcall lcd_cmd
	ldi r16, 2
	rcall milliwait1
	pop r16
	ret

lcd_home:
	push r16
	ldi r16, 0x02
	rcall lcd_cmd
	ldi r16, 2
	rcall milliwait1
	pop r16
	ret
lcd_on:	push r16
	ldi r16, 0x0E
	rcall lcd_cmd
	pop r16
	ret
lcd_off:
	push r16
	ldi r16, 0x08
	rcall lcd_cmd
	pop r16
	ret

lcd_init:
	push r16
	ldi r16, 0	;PORTB=0; (ohne PullupWiderstaende)
;	ldi r16, 3	;PORTB=0; (3 = +PullupWiderstaende)
	out PORTB, r16
	ldi r16, 50
	rcall milliwait	;//Warten bis LCD-Controller gebootet hat
	sbi PORTB, 5	;PORTB |= 0x20; //4-Bit-Modus einschalten
	rcall microwait1
	sbi PORTB, 3	;PORTB |= (1<<3); //sbi(PORTB,3); //Enable-Impuls
	rcall microwait1
	cbi PORTB, 3	;PORTB &= ~(1<<3); //cbi(PORTB,3);
	rcall milliwait5
	ldi r16, 0x28	;lcd_cmd(0x28);
	rcall lcd_cmd	;  //Funktions-Set: 2 Zeilen, 5*7 Matrix, 4Bit
	rcall lcd_off
	rcall lcd_clear
	ldi r16, 0x06	;lcd_cmd(0x06); // Entry Mode
	rcall lcd_cmd
	rcall lcd_on
	ldi r16, 0x11
	sts lcdpos, r16
	pop r16
	ret

lcd_go:
	push r17
	mov r17, r16
	swap r16
	rcall lcd_goto
	pop r17
	ret
lcd_goto:		;lcd_goto(r16=row,r17=col)
	dec r16		;row--
	andi r16, 0x01	;row &= 0x01; //nur Zeile 0 oder Zeile 1 zulassen
	swap r16
 	add r16, r16		
 	add r16, r16	;row <<= 6; //Zeile nach Bit 6 bringen
	dec r17		;col--;
	andi r17, 0x0F	;col &= 0x0F; //nur Bereich 0 bis 15 zulassen
	or r16, r17	;char tmp=row|col; //Adresse bilden
	ori r16, 0x80	;tmp |= 0x80; //Cursor setzen
	rcall lcd_cmd	;lcd_cmd(tmp);
	ret

.ifdef TESTCODE
lcd_r1:		;Ziffer in r1, und wenn r2 gesetzt 'M', auf LCD anzeigen
	push r16
	push r17
	ldi r16, 1
	ldi r17, 1
	rcall lcd_goto
	mov r16, r1
	andi r16, 0x0F
	rcall lcd_ziffer
	ldi r16, ' '
	rcall lcd_write
	ldi r16, ' '
	tst r2			; Mittelwertflag gesetzt?
	breq _ohne		; nein-->
	ldi r16, 'M'
_ohne:	rcall lcd_write
	pop r17
	pop r16
	ret

lcd_r16dez:
	ldi r17, 10
	rjmp lcd_r16
lcd_r16hex:
	ldi r17, 16
lcd_r16:	;r16 auf LCD, Dezimal oder Hexadezimal ohne fuehrende Nullen
		;Eingabe: r16=Zahl, r17=10 oder r17=16 fuer Hexadezimal
	push r0
	push r2
	push r3
	push r16
	push r17
	push r20
	push XL
	push XH
	clr r3
	mov r2, r16
	mov r20, r17		;Basis (10 oder 16)
	ldi XL, low(lcdn0)
	ldi XH, high(lcdn0)
	ldi r17, 1		;Zaehler fuer Anzahl Ziffern
	rjmp lc16L2
lc16L1:	rcall divmodr20		;durch 10 dividieren (oder durch 16)
	st X+, r0
	inc r17
lc16L2:	cp r2, r20		;Zahl>=10 ?
	brcc lc16L1		;ja--> weiter dividieren
	mov r16, r2
	rcall lcd_ziffer
	rjmp lc16L4
lc16L3:	ld r16, -X
	rcall lcd_ziffer
lc16L4:	dec r17
	brne lc16L3
	pop XH
	pop XL
	rjmp popr20ret

lcd_r16r17:	;Zahl r16:r17 auf LCD anzeigen, Wenn r18==1 lcdpos erhoehen
	push r0
	push r2
	push r3
	push r16
	push r17
	push r20
	mov r3, r16
	mov r2, r17
	lds r16, lcdpos
	rcall lcd_go
lcd_L0:	ldi r20, 10
	rcall divmodr20
	sts lcdn0, r0
	rcall divmodr20
	sts lcdn1, r0
	rcall divmodr20
	sts lcdn2, r0
	rcall divmodr20
	sts lcdn3, r0
	rcall divmodr20
	mov r16, r0
	tst r16			; wenn fuehrende Ziffer keine 0 ist..
	breq lcd_L1
	ori r16, 0x30		; ..dann die Ziffer selbst..
	rjmp lcd_L2
lcd_L1:	ldi r16, 0x20		; ..sonst ' ' ausgeben.
lcd_L2:	rcall lcd_write
	lds r16, lcdn3
	rcall lcd_ziffer
	lds r16, lcdn2
	rcall lcd_ziffer
	lds r16, lcdn1
	rcall lcd_ziffer
	lds r16, lcdn0
	rcall lcd_ziffer
	tst r18			; r18==0 ?
	breq lcd_L4		; ja--> lcdpos nicht veraendern
	lds r16, lcdpos
	cpi r16, 0x11
	breq lcd_L11
	cpi r16, 0x18
	breq lcd_L18
	cpi r16, 0x21
	breq lcd_L21
	ldi r16, 0x11
	rjmp lcd_L3
lcd_L11:ldi r16, 0x18
	rjmp lcd_L3
lcd_L18:
.ifdef DEBUGGING
	ldi r16, 0x11
.else
	ldi r16, 0x21
.endif
	rjmp lcd_L3
lcd_L21:ldi r16, 0x28
lcd_L3:	sts lcdpos, r16
lcd_L4:
popr20ret:
	pop r20
	pop r17
	pop r16
popr3ret:	
	pop r3
	pop r2
	pop r0
	ret

;-----------------------------------------------------------------------------
; divmodr20
;  16Bit vorzeichenlose Zahl dividieren durch 8Bit vorzeichenlosen Teiler
; Eingabe: r3:r2 = Zahl (16Bit), r20 = Teiler (8Bit) (unveraendert)
; Ausgabe: r3:r2 = Resultat, r0 = Rest
;-----------------------------------------------------------------------------
divmodr20:
	push r17
	clr r0
	ldi r17, 16
divL1:	lsl r2
	rol r3
	rol r0
	cp r0, r20
	brcs divL2
	sub r0, r20
	inc r2
divL2:	dec r17
	brne divL1
	pop r17
	ret
.endif
