;-----------------------------------------------------------------------------
;* Zwei Modellbau-Servo gleichzeitig ansteuern
;* Version: 0.5 	letzte Aenderung: 13.4.2010
;* Prozessor: ATtiny13A, interner Oszillator 8MHz
;* Schaltschema: zweiservo.png
;*   PB0, PB1 direkt zu Steuerleitung der Servos verbunden
;*   ADC1, ADC3 Potis als Eingabe zum Austesten
;* Autor: Rolf Pfister
;* Copyright: Freeware
;* History:
;* 9.4.2010	Erstellung
;*
;* globale Registerbelegung:
;* r20 = ADC-Kanal
;* r0 = Sicherung Statusregister im AD-Wandler-Interrupt
;* r23:r22 = aktueller PWM-Wert im Timer
;* r25:r24 = PWM-Wert in 1usec-Schritten, (Servo Pulse von etwa 0.5 bis 2.5 ms)
;* r27:r26 = PWM-Wert in 1usec-Schritten fuer Servo2
;* r10...r15 Zeiten (msec bis Tage)
;* r9 = Zaehler im Timerinterrupt

;Werte in Microsekunden:
.equ MINPWM=500
.equ MITTEPWM=1500
.equ MAXPWM=2500
.equ MINPWM2=600
.equ MITTEPWM2=1500
.equ MAXPWM2=2400

;.equ FREQTEST=1
;.equ FREQTEST2=1
;.equ TIMERTEST=1
;.equ TEST1=1
;.equ TEST2=1

;-----------------------------------------------------------------------------
.include "tn13def.inc"		; Definitionsdatei fuer den Prozessortyp
; Reset und Interruptvectoren	;VNr. Beschreibung
begin:	rjmp	main		; 1   Power On Reset
	reti			; 2   IRQ0 Handler
	reti			; 3   PCINT0
	rjmp	tc0over		; 4   Timer0 Overflow
	reti			; 5   EEPROM ready
	reti			; 6   Analog Comperator
	reti			; 7   Timer0 CompareA
	rjmp	tc0compb	; 8   Timer0 CompareB
	reti			; 9   Watchdog Interrupt
	rjmp	adcomplete	; 10  ADC Conversion
;-----------------------------------------------------------------------------
; Variablen im SRAM:
.equ SRAMSTART=0x60
.equ motorstatus=SRAMSTART	;0=gestoppt, 1=aktiv
.equ motor2status=SRAMSTART+1	;0=gestoppt, 1=aktiv
.equ adc1wertL=SRAMSTART+2
.equ adc1wertH=SRAMSTART+3	;Eingelesener Wert von ADC1
.equ adc2wertL=SRAMSTART+4
.equ adc2wertH=SRAMSTART+5	;Eingelesener Wert von ADC2
.equ adc3wertL=SRAMSTART+6
.equ adc3wertH=SRAMSTART+7	;Eingelesener Wert von ADC3

;-----------------------------------------------------------------------------
; Start, Power On, Reset
main:	ldi r16, low(RAMEND)
	out SPL, r16  	     ; Init Stackpointer
	ldi r16, 0x80
	out CLKPR, r16		; Vorbereitung zum Prescaler setzen
	ldi r16, 0
	out CLKPR, r16		; Clk-Prescaler auf 1 setzen --> 9.6MHz

	ldi r16, 0x61	; Wert ausprobieren fuer 8MHz
;	ldi r16, 0x45		; Wert ausprobieren fuer 6.4MHz
	out OSCCAL, r16		; Oszillator kallibrieren

	ldi r16, 0x03		;PB0, PB1 auf Ausgang
        out DDRB, r16		;Data Direction Register B setzen
	ldi r16, 0
	out PORTB, r16		;alle Ausgaenge auf Null

.ifdef FREQTEST
T1:	inc r16			; Test fuer Frequenzmessung
	out PORTB, r16		; Schlaufe braucht 4 Takte, eine Periode also 8
	rjmp T1			; Bei 8MHz muesste an B0 1MHz zu messen sein.
				; gemessen: 1.012 MHz
.endif

;; Timer auf 1MHz (1usec) und 250 setzen: 250*1usec=0.25msec
;; bei jedem 80. Durchgang (alle 20msec) PWM neu starten
	ldi r17, 0x0A		;Interruptenable setzen: B=8 A=4 OVR=2
	out TIMSK0, r17
;	ldi r17, 2		;CTC-Modus
; Ohne benutzen von Pin OC0B:
	ldi r17, 3		;PWM-Modus 3 oder 7
; Wenn Pin OC0B (PB1 beim tiny13a) automatisch gesetzt werden soll:
;	ldi r17, (1<<COM0B1)+3	;OC0B benutzen und PWM-Modus 3 oder 7
	out TCCR0A, r17		;Timer-Modus setzen
; Fuer PWM-Modus 7:
	ldi r17, 2+8		;1=nicht teilen, 2=durch 8, 3=64, 4=256, 5=1024
; Fuer andere Moden:
;	ldi r17, 2		;1=nicht teilen, 2=durch 8, 3=64, 4=256, 5=1024
	out TCCR0B, r17		;Vorteiler auf 64 setzen (6.4MHz/64=100kHz)
	ldi r17, 249
	out OCR0A, r17		;TOP=249 fuer teilen durch 250
	ldi r17, 0
	out OCR0B, r17		;PWM-Wert loeschen
	ldi r17, 0
	out TCNT0, r17		;Counter auf Startwert setzen
	ldi r17, 1
	mov r9, r17		;Durchlaufzaehler setzen

	clr r10
	clr r11			; r11r10 = Millisec
	clr r12			; Sekunden
	clr r13			; Minuten
	clr r14			; Stunden
	clr r15			; Tage, auf 0 setzen
	sei			;Interrupt erlauben

; AD-Wandler initialisieren:
	ldi r20, 3		;mit ADCx beginnen
	ldi r16, 0		;Referenz=VCC
;	ldi r16, 0x40		;Referenz= etwa 1.1V
	add r16, r20
	out ADMUX, r16
;Prescaler: 3=8 4=16 ... 6=64, 8MHz/64=125kHz (muss im Bereich 50-200 liegen)
;AD-Wandlung braucht jeweils 13 Takte also etwa 104usec
.equ	PRESCALER=6
;r16 = Prescaler+0b11001000 (ADEN+ADSC+ADIE = Enable,Einzelstart,Interrupt)
;ADIF (0x10) wird automatisch vom AD-Wandler gesetzt wenn Wandlung fertig.
	ldi r16, 0b11001000 + PRESCALER
	out ADCSRA, r16		;Prescaler und Enable
		;ab hier wird also der AD-Wandler dauernd laufen

;PWM-Steuerung initialisieren:
	ldi r24, low(MITTEPWM)
	ldi r25, high(MITTEPWM) ;Mittelstellung
	ldi r16, 1
	sts motorstatus, r16
	ldi r26, low(MITTEPWM2)
	ldi r27, high(MITTEPWM2) ;Mittelstellung
	ldi r16, 1
	sts motor2status, r16

.ifdef FREQTEST2
T1:	rjmp T1
.endif

.ifdef TIMERTEST
T1:	inc r18
	out PORTB, r18
	ldi r16, 100
	rcall milliwait1	;gemessen: 501.8Hz (mit 100: 50.2Hz)
	rjmp T1
.endif

;-----------------------------------------------------------------------------
;; Hauptprogramm
.ifdef TEST1
mainloop:
	ldi r24, low(MITTEPWM)
	ldi r25, high(MITTEPWM)
	ldi r16, 3
	rcall secwait		;3 sec warten
t1:	ldi r16, 1
	rcall secwait		;langsam von mittlerer ...
	subi r24, low(-10)
	sbci r25, high(-10)
	ldi r16, low(MAXPWM)	;auf maximale Position fahren.
	ldi r17, high(MAXPWM)	;auf maximale Position fahren.
	cp r24, r16
	cpc r25, r17
	brcs t1

	ldi r24, low(MITTEPWM)
	ldi r25, high(MITTEPWM)
	ldi r16, 3
	rcall secwait		;3 sec warten
	
t2:	ldi r16, 1
	rcall secwait		;langsam von mittlerer ...
	subi r24, 10
	sbci r25, 0
	ldi r16, low(MINPWM)	;auf minimale Position fahren.
	ldi r17, high(MINPWM)	;auf minimale Position fahren.
	cp r24, r16
	cpc r25, r17
	brcc t2
	rjmp mainloop

testloop:
	ldi r24, low(MAXPWM)
	ldi r25, high(MAXPWM)	;Maximal-Position
	ldi r16, 3
	rcall secwait		;3 Sekunden laufen lassen
	ldi r24, low(MINPWM)
	ldi r25, high(MINPWM)	;Minimal-Position
	ldi r16, 3
	rcall secwait		;3 Sekunden laufen lassen
	ldi r24, low((MINPWM+MAXPWM)/2)
	ldi r25, high((MINPWM+MAXPWM)/2)  ;Mittel-Position
	ldi r16, 3
	rcall secwait		;3 Sekunden laufen lassen
	rjmp testloop
.else
	rcall wait2einhalb
mainloop:
.ifdef TEST2
	ldi r24, low(880)	;Servo stellen: 45Grad = Ausposition
	ldi r25, high(880)	;Servo stellen: 45Grad = Ausposition
	lds r17, adc1wertH
	lds r16, adc1wertL	;r17:r16=adwert
	ldi r18, 3
	rcall div18		; auf Bereich 0...128 verkleinern
	cpi r16, 96		; Obere Poti-Stellung ?
	brcc minus		; ja-->
	cpi r16, 32		; Untere Poti-Stellung ?
	brcs plus		; ja-->
mitte:				;Poti ungefaehr in Mitte: automatische Sequenz
	mov r16, r12
	andi r16, 0x0F
	brne mainloop		; ca 16 Sec auf Aus-Position bleiben
	ldi r24, low(1620)	; 135Grad
	ldi r25, high(1620)	; 135Grad
	rcall wait2einhalb	; 2.5 Sec auf neuer Position
	rjmp mainloop
plus:
	ldi r24, low(1620)	; 135Grad
	ldi r25, high(1620)	; 135Grad
	rcall wait2einhalb	; 2.5 Sec auf neuer Position
	ldi r24, low(880)	; Servo auf Aus-Position
	ldi r25, high(880)	; Servo auf Aus-Position
	ldi r16, low(500)
	ldi r17, high(500)
	rcall milliwait		; 0.5 Sec um Aus-Position zu erreichen
	rjmp mainloop
minus:
	rjmp mainloop
.endif

; Hauptprogramm: 2 Servos mit 2 Potis ansteuern
mainloop2:
	lds r17, adc1wertH
	lds r16, adc1wertL	;r17:r16=Potistellung (Wert von 0 bis 1023)
	rcall umrechnen
	mov r24, r16
	mov r25, r17		;Servo stellen
	lds r17, adc3wertH
	lds r16, adc3wertL	;r17:r16=Potistellung (Wert von 0 bis 1023)
	rcall umrechnen
	mov r26, r16
	mov r27, r17		;Servo2 stellen
;	ldi r16, 50
;	rcall milliwait1	;Kurze Pause (ev. nicht noetig)
	rjmp mainloop2
umrechnen:
	add r16, r16
	rol r17			; *2 --> Wert von 0 bis 2046
	ldi r18, low(MINPWM)
	ldi r19, high(MINPWM)
	add r16, r18
	adc r17, r19		; +500 --> Wert von 500 bis 2546
	ret
.endif

;; Kleine Unterprogramme:
div18:		;r17:r16 /= 2^r18 ,r18 zerstoert
	asr r17
	ror r16
	dec r18
	brne div18
	ret

wait2einhalb:			; 2.5 sec warten, r16r17 zerstoert
	ldi r16, low(500)
	ldi r17, high(500)
	rcall milliwait		; 0.5 Sec
	ldi r16, 2		; 2 Sec
secwait:	; r16 Sekunden lang warten
	push r16
	push r17
	push r18
	mov r18, r16
	add r18, r16
_sw1:	ldi r16, low(500)
	ldi r17, high(500)
	rcall milliwait
	dec r18
	brne _sw1
	pop r18
	pop r17
	pop r16
	ret

;-----------------------------------------------------------------------------
; OVR-Interrupt-Routine, wird alle 0.25msec aufgerufen
; Registerbelegungen:
; r9=Durchlaufzaehler, jeder 80.Durchlauf den PWM-Wert setzen (80*0.25=20msec)
; r11:r10 Millisekunden, r12 Sekunden, r13 Minuten, r14 Stunden, r15 Tage
tc0over:push r16
	in r16, SREG	;Statusregister sichern
	push r16
	dec r9			;Durchlaufzaehler -= 1
.ifdef FREQTEST2
	out PORTB, r9		;gemessen: 2.001 kHz
	rjmp pop16reti
.endif
	mov r16, r9
	breq tc2		;wenn r9==0 --> Puls starten
	cpi r16, 40		;wenn r9==40 --> Puls fuer Servo2 starten
	breq tc22
	cpi r16, 1		;r9==1 ?
	brne tc1a2		;nein-->
	mov r22, r24		;ja: jeder 80. Durchlauf neuer PWM-Wert starten
	mov r23, r25		;r23:r22 = r25:r24
	rjmp tc1a
tc1a2:	cpi r16, 41	;r9==41 ?
	brne tc1a
	mov r22, r26	;ja: jeder 40. Durchlauf neuer PWM-Wert fuer Servo2
	mov r23, r27	;r23:r22 = r27:r26
tc1a:	tst r23
	brne tc1b		; >250 -->
	cpi r22, 250		; r23:r22 >= 250 ?
	brcc tc1b		; ja -->
	out OCR0B, r22	;PWM-Wert zetzen, ist erst beim naechsten mal aktiv
	clr r22
tcend:	tst r9
	breq tc3
	ldi r16, 3		;ist r9 durch 4
	and r16, r9		;teilbar?
	breq tc4		;ja--> Millsekunden raufzaehlen
pop16reti:
	pop r16
	out SREG, r16	;Statusregister zurueckholen
	pop r16
	reti
tc1b:	subi r22, 250
	sbci r23, 0		; r23:r22 -= 250
	ldi r16, 250
	out OCR0B, r16	;PWM-Wert zetzen, ist erst beim naechsten mal aktiv
	rjmp tcend
tc2:	lds r16, motorstatus
	tst r16			;Motor gestoppt?
	breq tc1a		;ja--> kein Puls senden
	sbi PORTB, 0		;Puls an PB0 starten
	rjmp tc1a
tc22:	lds r16, motor2status
	tst r16			;Motor gestoppt?
	breq tc1a		;ja--> kein Puls senden
	sbi PORTB, 1		;Puls an PB1 starten
	rjmp tc1a
tc3:	ldi r16, 80
	mov r9, r16
tc4:	sei
millisekzaehlen:
	push r17
	inc r10		;Millisekunden erhoehen
	brne L1
	inc r11
L1:	ldi r16, low(1000)
	ldi r17, high(1000)	;r17:r16 = 1000
	cp r10, r16
	cpc r11, r17
	brne end1	;fertig wenn msec!=1000
	clr r10
	clr r11		;Millisekunden auf 0 setzen
	inc r12		;Sekunden erhoehen
	ldi r16, 60
	cp r12, r16
	brne end1	;fertig wenn sec!=60
	clr r12		;Sekunden auf 0 setzen
	inc r13		;Minuten erhoehen
	cp r13, r16
	brne end1	;fertig wenn min!=60
	clr r13		;Minuten auf 0 setzen
	inc r14		;Stunden erhoehen
	ldi r16, 24
	cp r14, r16
	brne end1	;fertig wenn std!=24
	clr r14		;Stunden auf 0 setzen
	inc r15		;Tage erhoehen
end1:	pop r17
	rjmp pop16reti

tc0compb:			;Ende des PWM-Pulses
	cbi PORTB, 0
	cbi PORTB, 1
	reti

;-----------------------------------------------------------------------------
; Interrupt-Routine fuer AD-Wandler
; Rueckgabewert: adc1wert, adc2wert, adc3wert
adcomplete:
	push r16
	push r17
	in r17, ADCL	;L muss zuerst gelesen werden!
	in r16, ADCH
	sei		;Interrupt erlauben (fuer schnelles PWM)
	in r0, SREG	;Statusregister sichern
	cpi r20, 3
	breq _ad3
	cpi r20, 2
	breq _ad2
_ad1:	sts adc1wertL, r17 	;Messwert im RAM
	sts adc1wertH, r16	;zwischenspeichern.
	rjmp _ad4
_ad2:	sts adc2wertL, r17 	;Messwert im RAM
	sts adc2wertH, r16		;zwischenspeichern.
	rjmp _ad4
_ad3:	sts adc3wertL, r17 	;Messwert im RAM
	sts adc3wertH, r16	;zwischenspeichern.
_ad4:
	dec r20		;automatisch naechster AD-Kanal
	brne _ad5
	ldi r20, 3		;ADC3 als Anfangswert
_ad5:	ldi r16, 0		;Referenz=VCC
;	ldi r16, 0x40		;Referenz= etwa 1.1V
	add r16, r20		;ADC-Kanal
	out ADMUX, r16
	sbi ADCSRA, ADSC	;Starte naechste AD-Wandlung
	out SREG, r0	;Statusregister zurueckholen
	pop r17
	pop r16
	ret

;-----------------------------------------------------------------------------
; milliwait  optimierte Variante  Wartezeit 1 bis 1000 Millisekunden
; in r11:r10 wird durch einen Timerinterrupt die aktuelle Millisekunde
; gespeichert (16Bit-Zahl von 0 bis 999).
; Die aktelle Sekunde ist in r12 (wird hier noch nicht benutzt, da nur
; bis maximal 1000ms gewartet werden soll).
milliwait1:			;Wartezeit: r16, 1 bis 255 msec
	push r17
	clr r17
	rjmp miL1
milliwait:			;Wartezeit: r17:r16 msec, 1 bis 1000
	push r17
miL1:	push r16
	push r18
	mov r18, r10
milop1:	cp  r18, r10
	breq milop1		;1. Millisekunde abwarten (0 bis 1 ms)
	add r16, r10
	adc r17, r11		;sollzeit = Wartezeit + aktuelle Millisek
	subi r16, low(1001)
	sbci r17, high(1001)	;sollzeit -= 1; sollzeit -= 1000;
	brpl milop2		;wenn sollzeit>=0 -->
	subi r16, low(-1000)	;sonst haben wir 1000 zuviel subtrahiert
	sbci r17, high(-1000)	;also wieder 1000 addieren (soll -= -1000)
milop2:	cp r16, r10
	cpc r17, r11		;auf sollzeit warten
	brne milop2
	cp r16, r10		;falls zwischen 1. u 2. cp ein Interrupt war
	brne milop2
	pop r18
	pop r16
	pop r17
	ret
;-----------------------------------------------------------------------------
