;-----------------------------------------------------------------------------
;* LCD ueber nur 1 Datenleitung ansteuern
;* (mit GND und +5V gibt das dann 3 Leitungen)
;* Version: 0.2 	                   letzte Aenderung: 2.5.2010
;* Prozessor: ATtiny13A, interner Oszillator 8MHz
;* (sollte mit nur wenigen Aenderungen auch auf andern gehen)
;* Schaltschema: dreidrahtlcd.png
;*   PB4 als Ausgang zum LCD (Eingang zum ATtiny2313A auf dem LCD-Modul)
;*   ADC1 Poti als Eingabe zum Austesten
;* Autor: Rolf Pfister
;* Copyright: Freeware
;* History:
;* 18.4.2010	Erstellung, Version 0.1
;* 2.5.10	Wartepause (2ms) nach Senden eines Zeichens
;*
;* globale Registerbelegung:
;* r11:r10 Millisekunden

;* r12...r15 Zeiten (sec bis Tage, fuer LCD-Ansteuerung nicht noetig)
.define UHR

;.equ FREQTEST=1
;.equ TIMERTEST=1
;.define TEST1
;.define SLEEPTEST	;zum Strom sparen
.define SCHNELL		;schnelle Datenuebertragung
.define RUECKSIGNAL	;mit Rueckmeldung vom LCD

;-----------------------------------------------------------------------------
.include "tn13def.inc"		; Definitionsdatei fuer den Prozessortyp
; Reset und Interruptvectoren	;VNr. Beschreibung
begin:	rjmp	main		; 1   Power On Reset
	reti			; 2   IRQ0 Handler
	reti			; 3   PCINT0
	rjmp	tc0over		; 4   Timer0 Overflow
	reti			; 5   EEPROM ready
	reti			; 6   Analog Comperator
	reti			; 7   Timer0 CompareA
	reti			; 8   Timer0 CompareB
	reti			; 9   Watchdog Interrupt
	rjmp	adcomplete	; 10  ADC Conversion
;-----------------------------------------------------------------------------
; Variablen im SRAM:
.equ SRAMSTART=0x60
.equ adc1wertL=SRAMSTART+0
.equ adc1wertH=SRAMSTART+1	;Eingelesener Wert von ADC1
.equ lcdpos=SRAMSTART+2		;LCD-Position
.equ lcdn0=SRAMSTART+3
.equ lcdn1=SRAMSTART+4
.equ lcdn2=SRAMSTART+5
.equ lcdn3=SRAMSTART+6		;Zahlen fuers LCD

;-----------------------------------------------------------------------------
; LCD-Ansteuerung ueber nur 1 Datenleitung
.equ LCDPORT=PORTB
.equ LCDPIN=4
.ifdef RUECKSIGNAL
.equ LCDDDR=DDRB
.equ LCDPINB=PINB
.endif

lcd_init:		;LCD initialisieren
	push r16
	push r17
	push r18
	sbi LCDPORT, LCDPIN
	ldi r16, low(500)
	ldi r17, high(500)	;sicherstellen dass LCD-Modul gebootet hat
	rcall milliwait
	ldi r16, 0x11
	rcall lcd_go
	ldi r18, 20
lcd_i1:	ldi r16, ' '		;Anzeige loeschen
	rcall lcd_write
	dec r18
	brne lcd_i1
	ldi r16, 0x21		;zweite Zeile
	rcall lcd_go
	ldi r18, 20
lcd_i2:	ldi r16, ' '		;Anzeige loeschen
	rcall lcd_write
	dec r18
	brne lcd_i2
	ldi r16, 0x11
	rcall lcd_go
	pop r18
	pop r17
	pop r16
	ret
	
lcd_ziffer:		; r16 als Ziffer auf LCD anzeigen (0 <= r16 <= 15)
	ori r16, 0x30
	cpi r16, 0x3A	; Hex-Ziffer (r16>9) ?
	brcs lcd_write	; nein-->
	subi r16, -7	; fuer Hex-Ziffern z.B. 3A --> 41=='A'

lcd_write:		; r16 = Zeichen auf LCD senden, r16 zerstoert
	push r17
	ldi r17, 0
	rjmp lcd_se
lcd_cmd:push r17	; r16 = Kommando an LCD senden, r16 zerstoert
	ldi r17, 1
lcd_se:	push r18
	push r19
	mov r18, r16
	mov r19, r17
	cbi LCDPORT, LCDPIN	; Start-Bit
	rcall lcd_wait2
	rcall lcd_wait1		; 3 Einheit langer Startpuls
	sbi LCDPORT, LCDPIN	; Ende des Start-Bits
	ldi r17, 9		; 9 Bits senden
w_L1:	rcall lcd_wait1		; 1Einheit auf High
	cbi LCDPORT, LCDPIN
	andi r19, 1
	breq bit0
bit1:	rcall lcd_wait1		; kurzer Puls kodiert eine 1
	rjmp w_L2
bit0:	rcall lcd_wait2		; langer Puls kodiert eine 0
w_L2:	sbi LCDPORT, LCDPIN	; Ende des Bits
	rol r18
	rol r19
	dec r17
	brne w_L1
.ifdef RUECKSIGNAL
	cbi LCDDDR, LCDPIN	;Pin als Eingang
	sbi LCDPORT, LCDPIN	;Pullup-Widerstand
	ldi r18, 100
w_L3:	rcall lcd_wait1
	dec r18
	breq w_L5		;Abbruch wenn keine Rueckmeldung kommt
	in r16, LCDPINB
	andi r16, (1<<LCDPIN)
	brne w_L3
w_L4:	in r16, LCDPINB
	andi r16, (1<<LCDPIN)
	breq w_L4
w_L5:	sbi LCDDDR, LCDPIN	;Pin als Ausgang
.else
	ldi r16, 2
	rcall milliwait1	; warten bis LCD wieder bereit
.endif
	pop r19
	pop r18
	pop r17
	ret

lcd_go:	push r17
	sts lcdpos, r16
	mov r17, r16
	swap r16
	rcall lcd_goto
	pop r17
	ret
lcd_goto:		;lcd_goto(r16=row,r17=col)
	dec r16		;row--
	andi r16, 0x01	;row &= 0x01; //nur Zeile 0 oder Zeile 1 zulassen
	swap r16
 	add r16, r16		
 	add r16, r16	;row <<= 6; //Zeile nach Bit 6 bringen
	dec r17		;col--;
	andi r17, 0x0F	;col &= 0x0F; //nur Bereich 0 bis 15 zulassen
	or r16, r17	;char tmp=row|col; //Adresse bilden
	ori r16, 0x80	;tmp |= 0x80; //Cursor setzen
	rcall lcd_cmd	;lcd_cmd(tmp);
	ret

lcd_wait2:	;Warteschlaufe fuer lange Pulse
	rcall lcd_wait1
lcd_wait1:	;Warteschlaufe fuer kurze Pulse
.ifdef SCHNELL
;	push r17
;	ldi r17, 1
;w0:	ldi r16, 255
	ldi r16, 80
w1:	dec r16		;Schlaufe braucht 3 Takte
	brne w1		;70*3/6.4MHz = 32usec
;	dec r17		;255*3/6.4MHz = 120usec , 7*120usec = 0.84msec
;	brne w0
;	pop r17
.else
	ldi r16, 1
	rcall milliwait1
.endif
	ret

;-----------------------------------------------------------------------------
; Start, Power On, Reset
main:	ldi r16, low(RAMEND)
	out SPL, r16  	     ; Init Stackpointer
	ldi r16, 0x80
	out CLKPR, r16		; Vorbereitung zum Prescaler setzen
	ldi r16, 0
	out CLKPR, r16		; Clk-Prescaler auf 1 setzen --> 9.6MHz

;	ldi r16, 0x61		; Wert ausprobieren fuer 8MHz
	ldi r16, 0x4E		; Wert ausprobieren fuer 6.4MHz
	out OSCCAL, r16		; Oszillator kallibrieren

	ldi r16, (1<<4)		;PB4 auf Ausgang
        out DDRB, r16		;Data Direction Register B setzen
	ldi r16, (1<<4)
	out PORTB, r16		;Ausgang auf High

.ifdef FREQTEST
T1:	sbi PORTB, 4		; Test fuer Frequenzmessung
	nop
	nop
	cbi PORTB, 4		; Schlaufe braucht 8 Takte
	rjmp T1		;Bei 6.4MHz sollten 0.8 MHz an B4 zu messen sein
.endif

;; Timer auf 100kHz und Teiler 100 setzen: Interrupt jede Millisekunde
	ldi r17, 2		;Interruptenable setzen: B=8 A=4 OVR=2
	out TIMSK0, r17
	ldi r17, 0		;Normal-Modus
	out TCCR0A, r17		;Timer-Modus setzen
	ldi r17, 3		;1=nicht teilen, 2=durch 8, 3=64, 4=256, 5=1024
	out TCCR0B, r17		;Vorteiler auf 64 setzen (6.4MHz/64=100kHz)
.equ ZAEHLER=256-100
	ldi r17, ZAEHLER
	out TCNT0, r17		;Counter auf Startwert setzen

	clr r10
	clr r11			; r11r10 = Millisec
.ifdef UHR
	clr r12			; Sekunden
	clr r13			; Minuten
	clr r14			; Stunden
	clr r15			; Tage, auf 0 setzen
.endif
	sei			;Interrupt erlauben

; AD-Wandler initialisieren:
	ldi r20, 1		;mit ADCx beginnen
	ldi r16, 0		;Referenz=VCC
;	ldi r16, 0x40		;Referenz= etwa 1.1V
	add r16, r20
	out ADMUX, r16
;Prescaler: 3=8 4=16 ... 6=64, 6.4MHz/64=100kHz (muss im Bereich 50-200 liegen)
;AD-Wandlung braucht jeweils 13 Takte also etwa 130usec
.equ	PRESCALER=6
;r16 = Prescaler+0b11001000 (ADEN+ADSC+ADIE = Enable,Einzelstart,Interrupt)
;ADIF (0x10) wird automatisch vom AD-Wandler gesetzt wenn Wandlung fertig.
	ldi r16, 0b11001000 + PRESCALER
	out ADCSRA, r16		;Prescaler und Enable
		;ab hier wird also der AD-Wandler dauernd laufen

.ifdef TIMERTEST
T1:	sbi PORTB, 4
	ldi r16, 1
	rcall milliwait1
	cbi PORTB, 4
	ldi r16, 1
	rcall milliwait1	;sollte 500Hz geben
	rjmp T1
.endif

	rcall lcd_init

;-----------------------------------------------------------------------------
;; Hauptprogramm
mainloop:
;	ldi r16, 10
;	rcall secwait		;test
.ifdef TEST1
	ldi r16, 8
	rcall milliwait1
	ldi r16, 0x30		;'0'
	rcall lcd_write
	rjmp mainloop
.endif
	ldi r16, 100
	rcall milliwait1
	ldi r16, 0x11
	sts lcdpos, r16
	lds r16, adc1wertH
	lds r17, adc1wertL	;r16:r17=Potistellung (Wert von 0 bis 1023)
	ldi r18, 1
	rcall lcd_r16r17
	lds r16, adc1wertH
	lds r17, adc1wertL	;r16:r17=Potistellung (Wert von 0 bis 1023)
	clr r18
	rcall lcd_r16r17
.ifdef UHR
	ldi r16, 0x21		;auf 2.Zeile 1.Zeichen gehen
	rcall lcd_go
	mov r16, r14		;Stunden
	rcall lcd_r16
	ldi r16, ':'
	rcall lcd_write
	mov r16, r13		;Minuten
	rcall lcd_r16
	ldi r16, ':'
	rcall lcd_write
	mov r16, r12		;Sekunden
	rcall lcd_r16
	ldi r16, ' '
	rcall lcd_write
.endif
	rjmp mainloop

lcd_r16:		;r16 als 2 Ziffern auf LCD (inkl fuehrende 0)
	push r17
	rcall div10		;r16 /= 10, r17=Rest
	rcall lcd_ziffer	;erste Ziffer
	mov r16, r17
	rcall lcd_ziffer	;zweite Ziffer
	pop r17
	ret
	
lcd_r16r17:	;Zahl r16:r17 auf LCD anzeigen, Wenn r18==1 lcdpos erhoehen
	push r0
	push r2
	push r3
	push r16
	push r17
	push r20
	mov r3, r16
	mov r2, r17
	lds r16, lcdpos
	rcall lcd_go
lcd_L0:	ldi r20, 10
	rcall divmodr20
	sts lcdn0, r0
	rcall divmodr20
	sts lcdn1, r0
	rcall divmodr20
	sts lcdn2, r0
	rcall divmodr20
	sts lcdn3, r0
	rcall divmodr20
	mov r16, r0
	tst r16			; wenn fuehrende Ziffer keine 0 ist..
	breq lcd_L1
	ori r16, 0x30		; ..dann die Ziffer selbst..
	rjmp lcd_L2
lcd_L1:	ldi r16, 0x20		; ..sonst ' ' ausgeben.
lcd_L2:	rcall lcd_write
	lds r16, lcdn3
	rcall lcd_ziffer
	lds r16, lcdn2
	rcall lcd_ziffer
	lds r16, lcdn1
	rcall lcd_ziffer
	lds r16, lcdn0
	rcall lcd_ziffer
	tst r18			; r18==0 ?
	breq lcd_L4		; ja--> lcdpos nicht veraendern
	lds r16, lcdpos
	cpi r16, 0x11
	breq lcd_L11
	cpi r16, 0x18
	breq lcd_L18
	cpi r16, 0x21
	breq lcd_L21
	ldi r16, 0x11
	rjmp lcd_L3
lcd_L11:ldi r16, 0x18
	rjmp lcd_L3
lcd_L18:
.ifdef DEBUGGING
	ldi r16, 0x11
.else
	ldi r16, 0x21
.endif
	rjmp lcd_L3
lcd_L21:ldi r16, 0x28
lcd_L3:	sts lcdpos, r16
lcd_L4:
popr20ret:
	pop r20
	pop r17
	pop r16
popr3ret:	
	pop r3
	pop r2
	pop r0
	ret

div10:			;r16=r16/10, r17=Rest
	push r0
	push r2
	push r3
	push r20
	mov r2, r16
	clr r3
	ldi r20, 10
	rcall divmodr20
	mov r17, r0
	mov r16, r2
	pop r20
	pop r3
	pop r2
	pop r0
	ret

;-----------------------------------------------------------------------------
; divmodr20
;  16Bit vorzeichenlose Zahl dividieren durch 8Bit vorzeichenlosen Teiler
; Eingabe: r3:r2 = Zahl (16Bit), r20 = Teiler (8Bit) (unveraendert)
; Ausgabe: r3:r2 = Resultat, r0 = Rest
;-----------------------------------------------------------------------------
divmodr20:
	push r17
	clr r0
	ldi r17, 16
divL1:	lsl r2
	rol r3
	rol r0
	cp r0, r20
	brcs divL2
	sub r0, r20
	inc r2
divL2:	dec r17
	brne divL1
	pop r17
	ret

;-----------------------------------------------------------------------------
; OVR-Interrupt-Routine, wird alle msec aufgerufen
; Registerbelegungen:
; r11:r10 Millisekunden, r12 Sekunden, r13 Minuten, r14 Stunden, r15 Tage
tc0over:push r16
	in r16, SREG	;Statusregister sichern
	push r16
	ldi r16, ZAEHLER
	out TCNT0, r16	;Counter auf Startwert setzen
millisekzaehlen:
	push r17
	inc r10		;Millisekunden erhoehen
	brne L1
	inc r11
L1:	ldi r16, low(1000)
	ldi r17, high(1000)	;r17:r16 = 1000
	cp r10, r16
	cpc r11, r17
	brne end1	;fertig wenn msec!=1000
	clr r10
	clr r11		;Millisekunden auf 0 setzen
.ifdef UHR
	inc r12		;Sekunden erhoehen
	ldi r16, 60
	cp r12, r16
	brne end1	;fertig wenn sec!=60
	clr r12		;Sekunden auf 0 setzen
	inc r13		;Minuten erhoehen
	cp r13, r16
	brne end1	;fertig wenn min!=60
	clr r13		;Minuten auf 0 setzen
	inc r14		;Stunden erhoehen
	ldi r16, 24
	cp r14, r16
	brne end1	;fertig wenn std!=24
	clr r14		;Stunden auf 0 setzen
	inc r15		;Tage erhoehen
.endif
end1:	pop r17
pop16reti:
	pop r16
	out SREG, r16	;Statusregister zurueckholen
	pop r16
	reti

;-----------------------------------------------------------------------------
; Interrupt-Routine fuer AD-Wandler
; Rueckgabewert: adc1wert
adcomplete:
	push r16
	in r16, SREG	;Statusregister sichern
	push r16
	push r17
	in r17, ADCL	;L muss zuerst gelesen werden!
	in r16, ADCH
	sts adc1wertL, r17 	;Messwert im RAM
	sts adc1wertH, r16	;zwischenspeichern.
	ldi r17, 1		;ADC1 benutzen
	ldi r16, 0		;Referenz=VCC
;	ldi r16, 0x40		;Referenz= etwa 1.1V
	add r16, r17		;ADC-Kanal
	out ADMUX, r16
	sbi ADCSRA, ADSC	;Starte naechste AD-Wandlung
	pop r17
	rjmp pop16reti

;-----------------------------------------------------------------------------
; milliwait  optimierte Variante  Wartezeit 1 bis 1000 Millisekunden
; in r11:r10 wird durch einen Timerinterrupt die aktuelle Millisekunde
; gespeichert (16Bit-Zahl von 0 bis 999).
; Die aktelle Sekunde ist in r12 (wird hier noch nicht benutzt, da nur
; bis maximal 1000ms gewartet werden soll).
milliwait1:			;Wartezeit: r16, 1 bis 255 msec
	push r17
	clr r17
	rjmp miL1
milliwait:			;Wartezeit: r17:r16 msec, 1 bis 1000
	push r17
miL1:	push r16
	push r18
	mov r18, r10
milop1:	cp  r18, r10
	breq milop1		;1. Millisekunde abwarten (0 bis 1 ms)
	add r16, r10
	adc r17, r11		;sollzeit = Wartezeit + aktuelle Millisek
	subi r16, low(1001)
	sbci r17, high(1001)	;sollzeit -= 1; sollzeit -= 1000;
	brpl milop2		;wenn sollzeit>=0 -->
	subi r16, low(-1000)	;sonst haben wir 1000 zuviel subtrahiert
	sbci r17, high(-1000)	;also wieder 1000 addieren (soll -= -1000)
milop2:
.ifdef SLEEPTEST
	in r18, MCUCR
	ori r18, 0x20
	out MCUCR, r18
	sleep
	andi r18, ~0x20
	out MCUCR, r18
.endif
	cp r16, r10
	cpc r17, r11		;auf sollzeit warten
	brne milop2
	cp r16, r10		;falls zwischen 1. u 2. cp ein Interrupt war
	brne milop2
	pop r18
	pop r16
	pop r17
	ret
;-----------------------------------------------------------------------------
secwait:			;r16 Sekunden warten
	push r18
	mov r18, r16
sec_1:	ldi r16,low(1000)
	ldi r17, high(1000)
	rcall milliwait
	dec r18
	brne sec_1
	pop r18
	ret
