;* Sammlung diverser Funktionen/Unterprogramme in AVR-Assembler
;* Copyright: Freeware
;*
;History:
;20.5.2009	Erstellung
;
; Liste der bisher vorhandenen Unterprogramme
; -------------------------------------------
;Mathematische Funktionen:
; Divisionen: divmod, divmodu16u8, divmodu16u16, divmodu32u16
; Multiplikationen: mulu16
;Zeitbehandlungen:
; Timer-Uhrzeit: timer0_init, tc0over, timer1_init, timer1_match_A
; Wartefunktionen: milliwait
;
;-----------------------------------------------------------------------------
.include "m8def.inc" 	; Definitionsdatei fuer den Prozessortyp

;-----------------------------------------------------------------------------
; divmod  15Bit Vorzeichenlose Division
; Eingabe: r3:r2 = Bit-Zahl (15Bit), r21:r20 = Teiler (16Bit)
; Ausgabe: r21:r20 = Resultat, r3:r2 = Rest
; Speicherverbrauch: 74 Byte
; Anzahl Takte:	33 - 393
;  Schnell wenn Zahl klein und Teiler gross ist
;  Teiler muss groesser 0 sein, sonst haengt das Programm
;-----------------------------------------------------------------------------
divmod:	push r4
	push r5
	push r16
	push r17
	ldi r16, 1
	ldi r17, 0		;z=1
	clr r4
	clr r5			;Vorlaeufiges Resultat=0
divL1:	cp r2, r20
	cpc r3, r21		;teiler>zahl ?
	brcs divend1		;ja-->
	lsl r20
	rol r21			;teiler*=2
	lsl r16
	rol r17			;z*=2
	rjmp divL1
divend1:		;teiler ist jetzt groesser als zahl
	cpi r16, 1	;ist teiler noch verdoppelt?
	breq divend2	;nein --> fertig
	lsr r21
	ror r20			;teiler/=2
	lsr r17
	ror r16			;z/=2
	cp r2, r20
	cpc r3, r21		;teiler>zahl ?
	brcs divend1		;ja-->
	sub r2, r20
	sbc r3, r21		;nein: zahl-=teiler
	add r4, r16
	adc r5, r17		;resultat += z
	rjmp divend1
divend2:mov r20, r4
	mov r21, r5		;r21:r20=Resultat  ;r3:r2=Rest
	pop r17
	pop r16
	pop r5
	pop r4
	ret

;-----------------------------------------------------------------------------
; divmodu16u8
;  16Bit vorzeichenlose Zahl dividieren durch 8Bit vorzeichenlosen Teiler
; Eingabe: r3:r2 = Zahl (16Bit), r0 = Teiler (8Bit)
; Ausgabe: r3:r2 = Resultat, r0 = Rest
; Speicherverbrauch: 34 Byte
; Anzahl Takte: 158 - 174
;  Wenn Teiler 0 ist wird das Resultat 0xFFFF sein
;-----------------------------------------------------------------------------
divmodu16u8:
	push r16
	push r17
	mov r16, r0
	clr r0
	ldi r17, 16
L1:	lsl r2
	rol r3
	rol r0
	cp r0, r16
	brcs L2
	sub r0, r16
	inc r2
L2:	dec r17
	brne L1
	pop r17
	pop r16
	ret

;-----------------------------------------------------------------------------
; divmodu16u16
;  16Bit vorzeichenlose Zahl dividieren durch 16Bit vorzeichenlosen Teiler
; Eingabe: r3:r2 = Zahl, r1:r0 = Teiler
; Ausgabe: r3:r2 = Resultat, r1:r0 = Rest
; Speicherverbrauch: 48 Byte
; Anzahl Takte: 196 - 228
;  Wenn Teiler 0 ist wird das Resultat 0xFFFF sein
;-----------------------------------------------------------------------------
divmodu16u16:
	push r16
	push r17
	push r18
	mov r16, r0
	mov r17, r1
	clr r0
	clr r1
	ldi r18, 16
L3:	lsl r2
	rol r3
	rol r0
	rol r1
	cp r0, r16
	cpc r1, r17
	brcs L4
	sub r0, r16
	sbc r1, r17
	inc r2
L4:	dec r18
	brne L3
	pop r18
	pop r17
	pop r16
	ret

;-----------------------------------------------------------------------------
; mulu16
;  Vorzeichenlose Multiplikation zweiter 16Bit-Zahlen
; Eingabe: r17:r16, r19:r18
; Ausgabe: r3:r2:r1:r0
mulu16:			;r3:r2:r1:r0 = r17:r16 * r19:r18
	push r4		;Abkuerzung fuer Kommentare:
	push r5		; AB=r17:r16, BC=r19:r18, EFGH=Ergebnis
	push r17
	mul r17, r19	;A*C
	mov r2, r0	; EFGH = AB*CD --> EF=A*C GH=B*D EFG+=A*D+B*C
	mov r3, r1	; r3:r2 = EF=A*C
	mul r16, r18	;B*D
	mov r4, r0
	mov r5, r1	; r5:r4 = GH=B*D
	mul r17, r18	;A*D
	clr r17
	add r5, r0	; EFG+=A*D
	adc r2, r1
	adc r3, r17
	mul r16, r19	;B*C
	add r5, r0	; EFG+=B*C
	adc r2, r1
	adc r3, r17
	mov r0, r4
	mov r1, r5
	pop r17
	pop r5
	pop r4
	ret

;-----------------------------------------------------------------------------
; divmodu32u16
; Vorzeichenlose Division einer 32Bit-Zahl durch eine 16Bit-Zahl
; Eingabe: r3:r2:r1:r0 / r17:r16
; Ausgabe: r3:r2:r1:r0 = Ergebnis, r17:r16 = Rest
divmodu32u16:			;r3:r2:r1:r0 /= r17:r16, r17:r16 = Rest
	push r18
	push r19
	push r20
	clr r18
	clr r19
	ldi r20, 32
div16L1:lsl r0
	rol r1
	rol r2
	rol r3
	rol r18
	rol r19
	cp  r18, r16
	cpc r19, r17
	brcs div16L2
	sub r18, r16
	sbc r19, r17
	inc r0
div16L2:dec r20
	brne div16L1
	mov r16, r18
	mov r17, r19		;r17:r16 = Rest
	pop r20
	pop r19
	pop r18
	ret

;-----------------------------------------------------------------------------
; Timer-Uhr
; Registerbelegungen:
; r11:r10 Millisekunden, r12 Sekunden, r13 Minuten, r14 Stunden, r15 Tage
.def millisecL=r10
.def millisecH=r11
.def sec=r12
.def min=r13
.def stunden=r14
.def tage=r15

; Timer0:
; Einschraenkungen: ist nur mit runder Anzahl MHz genau.
;	rjmp tc0over		; 10  TC0 Overflow
.equ	ZAEHLER =131	; 131 bei 8MHz, 190 bei 4.19MHz 198 oder 199 bei 3.68MHz
timer0_init:
	push r17
	ldi r17, 0x01
	out TIMSK, r17		;Timer/Counter0 Overflow einschalten
	ldi r17, 3		;1=nicht teilen, 2=durch 8, 3=64, 4=256, 5=1024
	out TCCR0, r17		;Vorteiler auf 64 setzen
	ldi r17, ZAEHLER
	out TCNT0, r17
	pop r17
	ret
; Timer0-Interrupt-Routine
tc0over:push r16
	push r17
	in r17, SREG	;Statusregister sichern
	push r17
	ldi r17, ZAEHLER
	out TCNT0, r17
	ldi r16, 0xE8
	ldi r17, 0x03	;r17:r16 = 1000 = 0x03E8
	inc r10		;Millisekunden erhoehen
	brne L1
	inc r11
L1:	cp r10, r16
	brne end1	;fertig wenn msec!=1000
	cp r11, r17
	brne end1	;fertig wenn msec!=1000
	clr r10
	clr r11		;Millisekunden auf 0 setzen
	inc r12		;Sekunden erhoehen
	ldi r16, 60
	cp r12, r16
	brne end1	;fertig wenn sec!=60
	clr r12		;Sekunden auf 0 setzen
	inc r13		;Minuten erhoehen
	cp r13, r16
	brne end1	;fertig wenn min!=60
	clr r13		;Minuten auf 0 setzen
	inc r14		;Stunden erhoehen
	ldi r16, 24
	cp r14, r16
	brne end1	;fertig wenn std!=24
	clr r14		;Stunden auf 0 setzen
	inc r15		;Tage erhoehen
end1:	pop r17
	out SREG, r17	;Statusregister zurueckholen
	pop r17
	pop r16
	reti

; Timer1:
; Laeuft mit beliebiger Anzahl MHz sehr genau.
;	rjmp timer1_match_A	; 7   TC1 Compare Match A

timer1_init:
; Quarz moeglichst genau ausmessen und hier eintragen
; in Hertz und Tausendstel Hertz:

.equ Hz=3686400			; exakt 3.6864MHz
.equ MilliHz=0			; 3'686'400.000 Hz

;//Ausmessung-Beispiel: 3.6864MHz ist nach 20 Stunden 0.35 sec nachgegangen:
;// 3686400 / (3600*20) * 0.35 = 17.92  3686400-35.84=3686364.16
;.equ Hz=3686364
;.equ MilliHz=160

.equ Rest1 = Hz-Hz/1000*1000	;//jede Sekunde zu korrigierender Fehler
.equ Rest2 = MilliHz*6/100	;//jede Minute
.equ Rest3 = ((MilliHz*6-Rest2*100)*60+50)/100 ;//jede Stunde
.equ NormalerCompwert = Hz/1000-1

timer1_init:
	push r16
	push r17
	ldi r16, (1<<WGM12)|(1<<CS10)	;CTC-OCR1A-Modus, Prescaler=1
	out TCCR1B, r16
	ldi r16, low(NormalerCompwert)
	ldi r17, high(Normalercompwert)
	out OCR1AH, r17
	out OCR1AL, r16
	ldi r16, 0		;Startwert des Timers = 0
	out TCNT1H, r16
	out TCNT1L, r16
	ldi r16, 1<<OCIE1A	;Timer1 Interrupts bei Vergleichswert
	out TIMSK, r16
	clr millisecH
	clr millisecL
	clr sec
	clr min
	clr stunden
	clr tage
	pop r17
	pop r16
	ret

timer1_match_A:
	push r16
	push r17
	push r18		;uchar mehr;
	clr r18			;mehr=0
	inc millisecL		;++ms
	brne tc1_L1
	inc millisecH
tc1_L1: ldi r16, low(1000)
	ldi r17, high(1000)
	cp  millisecL, r16
	cpc millisecH, r17
	breq tc1_sec
	ldi r16, low(1000-Rest1)
	ldi r17, high(1000-Rest1)
	cp  millisecL, r16
	cpc millisecH, r17	;millisec>=1000-Rest1 ?
	brcs tc1_fertig		;nein-->
	ldi r18, 1		;ja: mehr=1
tc1_fertig:
	ldi r16, low(NormalerCompwert)
	ldi r17, high(Normalercompwert)
	add r16, r18
	brcc tc1_L2
	inc r17
tc1_L2:	out OCR1AH, r17
	out OCR1AL, r16
	pop r18
	pop r17
	pop r16
	reti
tc1_sec:			; Sekunden erhoehen
	clr millisecL
	clr millisecH
	inc sec
	ldi r17, 60
.if Rest2!=0
	ldi r16, 60-Rest2
	cp sec, r16		;if(++sec>=60-Rest2) mehr=1; else mehr=0;
	brcs tc1_L3
	ldi r18, 1
.endif
tc1_L3:	cp  sec, r17		;sec==60?
	brne tc1_fertig		;nein-->
	clr sec
	clr r18
	inc min
.if Rest3!=0
	ldi r16, 60-Rest3
	cp min, r16		;if(++min>=60-Rest3) mehr=1; else mehr=0;
	brcs tc1_L4
	ldi r18, 1
.endif
tc1_L4:	cp min, r17
	brne tc1_fertig
	clr min
	clr r18
	inc stunden
	ldi r17, 24
	cp stunden, r17
	brne tc1_fertig
	clr stunden
	inc tage		;{stunden=0; ++tage;}
	rjmp tc1_fertig

;-----------------------------------------------------------------------------
; Wartefunktionen
;nicht optimierte Variante
milliwait0:                      ;Wartezeit: r17:r16 msec, min 1 max 999
        push r16
        push r17
        push r18
        push r19
	cli
        mov r18, r10
        mov r19, r11            ;r19r18 = sollzeit = aktuelle Millisec
	sei
        add r18, r16
        adc r19, r17            ;sollzeit += Wartezeit
	ldi r16, low(1000)
        ldi r17, high(1000)
        cp  r18, r16
	cpc r18, r17		;auf 1000 testen
        brcs loop1              ;sollzeit<1000 --> weiter
	sub r18, r16		;sonst sollzeit>=1000 --> subtrahiere 1000
        sbc r19, r17
loop1:	cp r18, r10
        cpc r19, r11	;auf sollzeit warten
        brne loop1
	cp r18, r10	;falls zwischendurch ein Interrupt war:
        brne loop1	;koennte z.B. von 0x00FF auf 0x0100 gewechselt
        pop r19		;haben waerend wir auf 0x01FF warten.
        pop r18
        pop r17
        pop r16
        ret

;-----------------------------------------------------------------------------
; milliwait  optimierte Variante  Wartezeit 1 bis 1000 Millisekunden
; in r11:r10 wird durch einen Timerinterrupt die aktuelle Millisekunde
; gespeichert (16Bit-Zahl von 0 bis 999).
; Die aktelle Sekunde ist in r12 (wird hier noch nicht benutzt, da nur
; bis maximal 1000ms gewartet werden soll).
milliwait1:			;Wartezeit: r16, 1 bis 255 msec
	push r17
	clr r17
	rjmp miL1
milliwait:			;Wartezeit: r17:r16 msec, 1 bis 1000
	push r17
miL1:	push r16
	push r18
	mov r18, r10
milop1:	cp  r18, r10
	breq milop1		;1. Millisekunde abwarten (0 bis 1 ms)
	add r16, r10
	adc r17, r11		;sollzeit = Wartezeit + aktuelle Millisek
	subi r16, low(1001)
	sbci r17, high(1001)	;sollzeit -= 1; sollzeit -= 1000;
	brpl milop2		;wenn sollzeit>=0 -->
	subi r16, low(-1000)	;sonst haben wir 1000 zuviel subtrahiert
	sbci r17, high(-1000)	;also wieder 1000 addieren (soll -= -1000)
milop2:	cp r16, r10
	cpc r17, r11		;auf sollzeit warten
	brne milop2
	cp r16, r10		;falls zwischen 1. u 2. cp ein Interrupt war
	brne milop2
	pop r18
	pop r16
	pop r17
	ret
;-----------------------------------------------------------------------------
